#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2025 Fushan Wen <qydwhotmail@gmail.com>
# SPDX-License-Identifier: MIT

import sys

from jobnotificationinterface import JobNotificationControlInterface
from KCoreAddons import KJob
from KJobWidgets import KUiServerV2JobTracker
from PySide6.QtCore import QCoreApplication, QMetaObject, Qt, QTimer
from PySide6.QtGui import QGuiApplication


class TestJob(KJob):

    def __init__(self) -> None:
        super().__init__()

        self.setProperty("immediateProgressReporting", True)
        self.setCapabilities(KJob.Capability.Killable | KJob.Capability.Suspendable)

    def start(self) -> None:
        # Delay reporting total bytes to test BUG 501991
        self.startElapsedTimer()
        QTimer.singleShot(500, self, self.init_progress)

    def doKill(self) -> bool:
        QMetaObject.invokeMethod(QCoreApplication.instance(), "quit", Qt.QueuedConnection)
        return True

    def doSuspend(self) -> bool:
        return True

    def doResume(self) -> bool:
        return True

    def init_progress(self) -> None:
        self.setProperty("destUrl", "file:///home/foobar/test.txt")
        self.setTotalAmount(KJob.Unit.Bytes, 1024 * 1024 * 10)
        self.setProcessedAmount(KJob.Unit.Bytes, 1024 * 1024 * 1)
        self.setTotalAmount(KJob.Unit.Files, 1)

    def update_progress(self, new_bytes: int) -> None:
        if new_bytes > self.processedAmount(KJob.Unit.Bytes):
            self.emitSpeed(new_bytes - self.processedAmount(KJob.Unit.Bytes))
        self.setProcessedAmount(KJob.Unit.Bytes, new_bytes)
        if self.processedAmount(KJob.Unit.Bytes) == self.totalAmount(KJob.Unit.Bytes):
            self.emitResult()


if __name__ == "__main__":
    app = QGuiApplication(sys.argv)
    app.setDesktopFileName("org.kde.plasmashell")

    tracker = KUiServerV2JobTracker(app)
    job = TestJob()
    job.finished.connect(app.quit)
    tracker.registerJob(job)
    job.start()

    control_interface = JobNotificationControlInterface(job)

    sys.exit(app.exec())



# SPDX-FileCopyrightText: 2026 Fushan Wen <qydwhotmail@gmail.com>
# SPDX-License-Identifier: MIT

"""Base test class for Appium tests."""

from __future__ import annotations

import os
import subprocess
import time
import unittest
from typing import ClassVar

from appium import webdriver
from appium.options.common.base import AppiumOptions

from .constants import APPIUM_SERVER_URL, DEFAULT_IMPLICIT_TIMEOUT, KDE_VERSION, PROCESS_WAIT_INTERVAL, PROCESS_WAIT_ITERATIONS


def _filter_fatal_warnings(environ: dict[str, str]) -> dict[str, str]:
    """Only enable QT_FATAL_WARNINGS when KDECI_BUILD is set."""
    if "QT_FATAL_WARNINGS" in environ and "KDECI_BUILD" not in os.environ:
        environ = dict(environ)
        del environ["QT_FATAL_WARNINGS"]
    return environ


class PlasmaAppletTest(unittest.TestCase):
    """Base test class for Plasma applet tests using plasmawindowed."""

    driver: ClassVar[webdriver.Remote]
    widget_id: ClassVar[str]
    extra_environ: ClassVar[dict[str, str]] = {}

    @classmethod
    def setUpClass(cls) -> None:
        options = AppiumOptions()
        options.set_capability("app", f"plasmawindowed -p org.kde.plasma.nano {cls.widget_id}")
        options.set_capability("timeouts", {'implicit': DEFAULT_IMPLICIT_TIMEOUT})
        environ: dict[str, str] = {"LC_ALL": "en_US.UTF-8"}
        environ.update(_filter_fatal_warnings(cls.extra_environ))
        options.set_capability("environ", environ)
        cls.driver = webdriver.Remote(command_executor=APPIUM_SERVER_URL, options=options)

    def tearDown(self) -> None:
        if not self._outcome.result.wasSuccessful():
            self.driver.get_screenshot_as_file(f"failed_test_shot_{self.widget_id}_#{self.id()}.png")

    @classmethod
    def tearDownClass(cls) -> None:
        subprocess.check_call([f"kquitapp{KDE_VERSION}", "plasmawindowed"])
        for _ in range(PROCESS_WAIT_ITERATIONS):
            try:
                subprocess.check_call(["pidof", "plasmawindowed"])
            except subprocess.CalledProcessError:
                break
            time.sleep(PROCESS_WAIT_INTERVAL)
        cls.driver.quit()


class KCMTest(unittest.TestCase):
    """Base test class for KCM tests using kcmshell."""

    driver: ClassVar[webdriver.Remote]
    kcm_id: ClassVar[str]
    extra_environ: ClassVar[dict[str, str]] = {}

    @classmethod
    def setUpClass(cls) -> None:
        options = AppiumOptions()
        options.set_capability("app", f"kcmshell{KDE_VERSION} {cls.kcm_id}")
        options.set_capability("timeouts", {'implicit': DEFAULT_IMPLICIT_TIMEOUT})
        environ: dict[str, str] = {"LC_ALL": "en_US.UTF-8"}
        environ.update(_filter_fatal_warnings(cls.extra_environ))
        options.set_capability("environ", environ)
        cls.driver = webdriver.Remote(command_executor=APPIUM_SERVER_URL, options=options)

    def tearDown(self) -> None:
        if not self._outcome.result.wasSuccessful():
            self.driver.get_screenshot_as_file(f"failed_test_shot_{self.kcm_id}_#{self.id()}.png")

    @classmethod
    def tearDownClass(cls) -> None:
        from appium.webdriver.common.appiumby import AppiumBy
        cls.driver.find_element(AppiumBy.XPATH, "//*[@name='Close' and contains(@accessibility-id, 'Button')]").click()
        for _ in range(PROCESS_WAIT_ITERATIONS):
            try:
                subprocess.check_call(["pidof", f"kcmshell{KDE_VERSION}"])
            except subprocess.CalledProcessError:
                break
            time.sleep(PROCESS_WAIT_INTERVAL)
        cls.driver.quit()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2019 The GNOME Music developers
# SPDX-FileCopyrightText: 2023 Fushan Wen <qydwhotmail@gmail.com>
# SPDX-License-Identifier: GPL-2.0-or-later

# pylint: disable=too-many-arguments

# For FreeBSD CI which only has python 3.9
from __future__ import annotations

import argparse
import logging
import json
import os
import signal
import sys
import threading
import time
from os import getpid
from typing import Any, Final

current_folder: Final = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(current_folder, os.pardir, os.pardir, "utils"))

from gi.repository import Gio, GLib
from GLibMainLoopThread import GLibMainLoopThread


def read_player_metadata(json_dict: dict[str, Any]) -> list[dict[str, GLib.Variant]]:
    song_list: list[dict[str, int | str | list[str]]] = json_dict["metadata"]
    assert len(song_list) > 0
    _metadata: list[dict[str, GLib.Variant]] = []
    for song in song_list:
        song_dict: dict[str, GLib.Variant] = {
            "mpris:trackid": GLib.Variant('o', song["mpris:trackid"]),
            "xesam:url": GLib.Variant('s', "file://" + os.path.join(current_folder, song["xesam:url"])),
            "mpris:length": GLib.Variant('x', int(song["mpris:length"])),  # ms
            "xesam:title": GLib.Variant('s', song["xesam:title"]),
            "xesam:artist": GLib.Variant('as', song["xesam:artist"]),
            "mpris:artUrl": GLib.Variant('s', "")
        }
        if "xesam:album" in song.keys():
            song_dict["xesam:album"] = GLib.Variant('s', song["xesam:album"])
        if "mpris:artUrl" in song.keys():
            song_dict["mpris:artUrl"] = GLib.Variant('s', "file://" + os.path.join(current_folder, song["mpris:artUrl"]))
        if "kde:pid" in song.keys():
            song_dict["kde:pid"] = GLib.Variant('u', song["kde:pid"])
        _metadata.append(song_dict)

    return _metadata


def read_base_properties(json_dict: dict[str, Any]) -> dict[str, GLib.Variant]:
    prop_dict: dict[str, bool | str | list[str]] = json_dict["base_properties"]
    assert len(prop_dict) > 0
    _base_properties: dict[str, GLib.Variant] = {
        "CanQuit": GLib.Variant('b', prop_dict["CanQuit"]),
        "Fullscreen": GLib.Variant('b', prop_dict["Fullscreen"]),
        "CanSetFullscreen": GLib.Variant('b', prop_dict["CanSetFullscreen"]),
        "CanRaise": GLib.Variant('b', prop_dict["CanRaise"]),
        "HasTrackList": GLib.Variant('b', prop_dict["HasTrackList"]),
        "Identity": GLib.Variant('s', prop_dict["Identity"]),
        "DesktopEntry": GLib.Variant('s', prop_dict["DesktopEntry"]),
        "SupportedUriSchemes": GLib.Variant('as', prop_dict["SupportedUriSchemes"]),
        "SupportedMimeTypes": GLib.Variant('as', prop_dict["SupportedMimeTypes"]),
    }
    return _base_properties


def read_player_properties(json_dict: dict[str, Any], _metadata: dict[str, GLib.Variant]) -> dict[str, GLib.Variant]:
    prop_dict: dict[str, int | float | bool | str | dict] = json_dict["player_properties"]
    assert len(prop_dict) > 0
    _player_properties: dict[str, GLib.Variant] = {
        "PlaybackStatus": GLib.Variant("s", prop_dict["PlaybackStatus"]),
        "LoopStatus": GLib.Variant("s", prop_dict["LoopStatus"]),
        "Rate": GLib.Variant("d", prop_dict["Rate"]),
        "Shuffle": GLib.Variant("b", prop_dict["Shuffle"]),
        "Metadata": GLib.Variant("a{sv}", _metadata),
        "Position": GLib.Variant("x", prop_dict["Position"]),
        "MinimumRate": GLib.Variant("d", prop_dict["MinimumRate"]),
        "MaximumRate": GLib.Variant("d", prop_dict["MaximumRate"]),
        "Volume": GLib.Variant("d", prop_dict["Volume"]),
        "CanGoNext": GLib.Variant("b", prop_dict["CanGoNext"]),
        "CanGoPrevious": GLib.Variant("b", prop_dict["CanGoPrevious"]),
        "CanPlay": GLib.Variant("b", prop_dict["CanPlay"]),
        "CanPause": GLib.Variant("b", prop_dict["CanPause"]),
        "CanSeek": GLib.Variant("b", prop_dict["CanSeek"]),
        "CanControl": GLib.Variant("b", prop_dict["CanControl"]),
    }
    return _player_properties


class Mpris2:
    """
    MPRIS2 interface implemented in GDBus, since dbus-python does not support property
    """

    BASE_IFACE: Final = "org.mpris.MediaPlayer2"
    OBJECT_PATH: Final = "/org/mpris/MediaPlayer2"
    PLAYER_IFACE: Final = GLib.Variant('s', "org.mpris.MediaPlayer2.Player")
    APP_INTERFACE: Final = f"org.mpris.MediaPlayer2.appiumtest.instance{str(getpid())}"

    connection: Gio.DBusConnection | None = None

    def __init__(self, metadata: list[dict[str, GLib.Variant]], base_properties: dict[str, GLib.Variant], player_properties: dict[str, GLib.Variant], current_index: int) -> None:
        self.metadata: list[dict[str, GLib.Variant]] = metadata
        self.base_properties: dict[str, GLib.Variant] = base_properties
        self.player_properties: dict[str, GLib.Variant] = player_properties
        self.current_index: int = current_index

        self.registered_event = threading.Event()
        self.playback_status_set_event = threading.Event()
        self.metadata_updated_event = threading.Event()

        self.__owner_id: int = Gio.bus_own_name(Gio.BusType.SESSION, self.APP_INTERFACE, Gio.BusNameOwnerFlags.NONE, self.on_bus_acquired, None, None)
        assert self.__owner_id > 0

        self.__prop_reg_id: int = 0
        self.__player_reg_id: int = 0
        self.__base_reg_id: int = 0

    def quit(self) -> None:
        if self.connection is None:
            return
        self.connection.unregister_object(self.__prop_reg_id)
        self.__prop_reg_id = 0
        self.connection.unregister_object(self.__player_reg_id)
        self.__player_reg_id = 0
        self.connection.unregister_object(self.__base_reg_id)
        self.__base_reg_id = 0
        Gio.bus_unown_name(self.__owner_id)
        self.connection.flush_sync(None)
        sync_event = threading.Event()
        GLib.idle_add(lambda: sync_event.set() or False)
        sync_event.wait(timeout=3.0)
        self._wait_for_dbus_name_released()
        logging.info("Player exit")

    def _wait_for_dbus_name_released(self, timeout: float = 5.0) -> None:
        """Wait for a D-Bus name to be released."""
        bus_name = self.APP_INTERFACE
        session_bus = Gio.bus_get_sync(Gio.BusType.SESSION)
        deadline = time.time() + timeout
        while time.time() < deadline:
            result = session_bus.call_sync("org.freedesktop.DBus", "/org/freedesktop/DBus", "org.freedesktop.DBus", "NameHasOwner", GLib.Variant("(s)", (bus_name, )), None, Gio.DBusSendMessageFlags.NONE, 100, None)
            if not result[0]:
                return
            time.sleep(0.1)

    def on_bus_acquired(self, connection: Gio.DBusConnection, name: str, *args) -> None:
        """
        Interface is ready, now register objects.
        """
        self.connection = connection

        dbus_folder: str = os.path.join(current_folder, os.pardir, os.pardir, os.pardir, "libkmpris", "dbus")

        with open(f"{dbus_folder}/org.freedesktop.DBus.Properties.xml", encoding="utf-8") as handler:
            properties_introspection_xml: str = '\n'.join(handler.readlines())
            introspection_data = Gio.DBusNodeInfo.new_for_xml(properties_introspection_xml)
            self.__prop_reg_id = connection.register_object(self.OBJECT_PATH, introspection_data.interfaces[0], self.properties_handle_method_call, None, None)
        assert self.__prop_reg_id > 0

        with open(f"{dbus_folder}/org.mpris.MediaPlayer2.Player.xml", encoding="utf-8") as handler:
            player_introspection_xml: str = '\n'.join(handler.readlines())
            introspection_data = Gio.DBusNodeInfo.new_for_xml(player_introspection_xml)
            self.__player_reg_id = connection.register_object(self.OBJECT_PATH, introspection_data.interfaces[0], self.player_handle_method_call, self.player_handle_get_property, self.player_handle_set_property)
        assert self.__player_reg_id != 0

        with open(f"{dbus_folder}/org.mpris.MediaPlayer2.xml", encoding="utf-8") as handler:
            interface_introspection_xml: str = '\n'.join(handler.readlines())
            introspection_data = Gio.DBusNodeInfo.new_for_xml(interface_introspection_xml)
            self.__base_reg_id = connection.register_object(self.OBJECT_PATH, introspection_data.interfaces[0], self.interface_handle_method_call, self.interface_handle_get_property, self.interface_handle_set_property)
        assert self.__base_reg_id != 0

        print("MPRIS registered", file=sys.stderr, flush=True)
        self.registered_event.set()

    def properties_handle_method_call(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, method_name: str, parameters: GLib.Variant, invocation: Gio.DBusMethodInvocation) -> None:
        """
        Handles method calls for org.freedesktop.DBus.Properties
        """
        assert interface_name == "org.freedesktop.DBus.Properties", f"Wrong interface name {interface_name} from {sender}"

        if method_name == "Get":
            ret: Any = None
            interface: str = parameters[0]
            property_name: str = parameters[1]
            if interface == self.BASE_IFACE:
                ret = self.base_properties[property_name]
            elif interface == self.PLAYER_IFACE.get_string():
                ret = self.player_properties[property_name]
            else:
                assert False, f"Unknown interface {interface}"

            print(f"Get: {interface} {property_name} {ret}", file=sys.stderr, flush=True)
            invocation.return_value(GLib.Variant.new_tuple(GLib.Variant('v', ret)))

        elif method_name == "GetAll":
            interface = parameters[0]
            if interface == self.BASE_IFACE:
                ret = GLib.Variant('a{sv}', self.base_properties)
            elif interface == self.PLAYER_IFACE.get_string():
                ret = GLib.Variant('a{sv}', self.player_properties)
            else:
                assert False, f"Unknown interface {interface}"

            print(f"GetAll: {interface} {ret}", file=sys.stderr, flush=True)
            invocation.return_value(GLib.Variant.new_tuple(ret))

        elif method_name == "Set":
            interface = parameters[0]
            property_name = parameters[1]
            value: Any = parameters[2]

            if interface != self.PLAYER_IFACE.get_string():
                assert False, f"Wrong interface {interface}"

            if property_name == "Rate":
                self.set_rate(value, connection, object_path)
            elif property_name == "LoopStatus":
                self.set_repeat(value, connection, object_path)
            elif property_name == "Shuffle":
                self.set_shuffle(value, connection, object_path)
            elif property_name == "Volume":
                self.set_volume(value, connection, object_path)
            else:
                assert False, f"Unknown property {property_name}"

            print(f"Set: {interface} {property_name} {value}", file=sys.stderr, flush=True)

        else:
            logging.error("Unhandled method: %s", method_name)
            invocation.return_error_literal(Gio.dbus_error_quark(), Gio.DBusError.UNKNOWN_METHOD, f"Unknown method {method_name}")

    def set_playing(self, playing: bool, connection: Gio.DBusConnection, object_path: str) -> None:
        """
        Changes the playing state
        """
        self.player_properties["PlaybackStatus"] = GLib.Variant('s', "Playing" if playing else "Paused")
        changed_properties = GLib.Variant('a{sv}', {
            "PlaybackStatus": self.player_properties["PlaybackStatus"],
        })
        Gio.DBusConnection.emit_signal(connection, None, object_path, "org.freedesktop.DBus.Properties", "PropertiesChanged", GLib.Variant.new_tuple(self.PLAYER_IFACE, changed_properties, GLib.Variant('as', ())))
        self.playback_status_set_event.set()

    def stop(self, connection: Gio.DBusConnection, object_path: str) -> None:
        """
        Changes the playing state
        """
        self.player_properties["PlaybackStatus"] = GLib.Variant('s', "Stopped")
        self.player_properties["Metadata"] = GLib.Variant('a{sv}', {})
        changed_properties = GLib.Variant('a{sv}', {
            "PlaybackStatus": self.player_properties["PlaybackStatus"],
            "Metadata": self.player_properties["Metadata"],
        })
        Gio.DBusConnection.emit_signal(connection, None, object_path, "org.freedesktop.DBus.Properties", "PropertiesChanged", GLib.Variant.new_tuple(self.PLAYER_IFACE, changed_properties, GLib.Variant('as', ())))

    def set_rate(self, rate: float, connection: Gio.DBusConnection, object_path: str) -> None:
        """
        Changes the shuffle state
        """
        assert isinstance(rate, float), f"argument is not a float but {type(rate)}"
        assert self.player_properties["MinimumRate"].get_double() <= rate <= self.player_properties["MaximumRate"].get_double(), f"Rate {rate} is out of bounds"
        self.player_properties["Rate"] = GLib.Variant('d', rate)
        changed_properties = GLib.Variant('a{sv}', {
            "Rate": self.player_properties["Rate"],
        })
        Gio.DBusConnection.emit_signal(connection, None, object_path, "org.freedesktop.DBus.Properties", "PropertiesChanged", GLib.Variant.new_tuple(self.PLAYER_IFACE, changed_properties, GLib.Variant('as', ())))

    def set_shuffle(self, shuffle: bool, connection: Gio.DBusConnection, object_path: str) -> None:
        """
        Changes the shuffle state
        """
        assert isinstance(shuffle, bool), f"argument is not a boolean but {type(shuffle)}"
        self.player_properties["Shuffle"] = GLib.Variant('b', shuffle)
        changed_properties = GLib.Variant('a{sv}', {
            "Shuffle": self.player_properties["Shuffle"],
        })
        Gio.DBusConnection.emit_signal(connection, None, object_path, "org.freedesktop.DBus.Properties", "PropertiesChanged", GLib.Variant.new_tuple(self.PLAYER_IFACE, changed_properties, GLib.Variant('as', ())))

    def set_repeat(self, repeat: str, connection: Gio.DBusConnection, object_path: str) -> None:
        """
        Changes the loop state
        """
        assert isinstance(repeat, str), f"argument is not a string but {type(repeat)}"
        self.player_properties["LoopStatus"] = GLib.Variant('s', repeat)
        changed_properties = GLib.Variant('a{sv}', {
            "LoopStatus": self.player_properties["LoopStatus"],
        })
        Gio.DBusConnection.emit_signal(connection, None, object_path, "org.freedesktop.DBus.Properties", "PropertiesChanged", GLib.Variant.new_tuple(self.PLAYER_IFACE, changed_properties, GLib.Variant('as', ())))

    def set_volume(self, volume: float, connection: Gio.DBusConnection, object_path: str) -> None:
        """
        Adjusts the volume
        """
        assert isinstance(volume, float) and 0 <= volume <= 1, f"Invalid volume {volume} of type {type(volume)}"
        self.player_properties["Volume"] = GLib.Variant('d', volume)
        changed_properties = GLib.Variant('a{sv}', {
            "Volume": self.player_properties["Volume"],
        })
        Gio.DBusConnection.emit_signal(connection, None, object_path, "org.freedesktop.DBus.Properties", "PropertiesChanged", GLib.Variant.new_tuple(self.PLAYER_IFACE, changed_properties, GLib.Variant('as', ())))

    def player_handle_method_call(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, method_name: str, parameters: GLib.Variant, invocation: Gio.DBusMethodInvocation) -> None:
        """
        Handles method calls for org.mpris.MediaPlayer2.Player
        """
        assert interface_name == "org.mpris.MediaPlayer2.Player", f"Wrong interface name {interface_name} from {sender}"

        print(f"player_handle_method_call method_name: {method_name}", file=sys.stderr, flush=True)

        if method_name == "Next" or method_name == "Previous":
            self.current_index += 1 if method_name == "Next" else -1
            assert 0 <= self.current_index < len(self.metadata)
            self.player_properties["Metadata"] = GLib.Variant('a{sv}', self.metadata[self.current_index])
            self.player_properties["CanGoPrevious"] = GLib.Variant('b', self.current_index > 0)
            self.player_properties["CanGoNext"] = GLib.Variant('b', self.current_index < len(self.metadata) - 1)
            changed_properties = GLib.Variant('a{sv}', {
                "Metadata": self.player_properties["Metadata"],
                "CanGoPrevious": self.player_properties["CanGoPrevious"],
                "CanGoNext": self.player_properties["CanGoNext"],
            })
            Gio.DBusConnection.emit_signal(connection, None, object_path, "org.freedesktop.DBus.Properties", "PropertiesChanged", GLib.Variant.new_tuple(self.PLAYER_IFACE, changed_properties, GLib.Variant('as', ())))
            self.metadata_updated_event.set()

        elif method_name == "Pause":
            self.set_playing(False, connection, object_path)

        elif method_name == "PlayPause":
            self.set_playing(self.player_properties["PlaybackStatus"].get_string() != "Playing", connection, object_path)

        elif method_name == "Stop":
            self.stop(connection, object_path)

        elif method_name == "Play":
            self.set_playing(True, connection, object_path)

        elif method_name == "Seek":
            offset: int = parameters[0]
            length: int = int(self.metadata[self.current_index]["mpris:length"])
            position: int = self.player_properties["Position"].get_int64()
            assert 0 <= position + offset <= length
            self.player_properties["Position"] = GLib.Variant('x', position + offset)
            changed_properties = GLib.Variant('a{sv}', {
                'Position': self.player_properties["Position"],
            })
            Gio.DBusConnection.emit_signal(connection, None, object_path, self.PLAYER_IFACE.get_string(), "Seeked", GLib.Variant.new_tuple(self.player_properties["Position"]))
            Gio.DBusConnection.emit_signal(connection, None, object_path, "org.freedesktop.DBus.Properties", "PropertiesChanged", GLib.Variant.new_tuple(self.PLAYER_IFACE, changed_properties, GLib.Variant('as', ())))

        elif method_name == "SetPosition":
            assert parameters[0] == self.metadata[self.current_index]["mpris:trackid"].get_string(), f"expected trackid: {parameters[0]}, actual trackid: {self.metadata[self.current_index]['mpris:trackid'].get_string()}"
            self.player_properties["Position"] = GLib.Variant('x', parameters[1])
            changed_properties = GLib.Variant('a{sv}', {
                'Position': self.player_properties["Position"],
            })
            Gio.DBusConnection.emit_signal(connection, None, object_path, "org.freedesktop.DBus.Properties", "PropertiesChanged", GLib.Variant.new_tuple(self.PLAYER_IFACE, changed_properties, GLib.Variant('as', ())))

        elif method_name == "OpenUri":
            print("OpenUri", file=sys.stderr, flush=True)

        else:
            # In case the interface adds new methods, fail here for easier discovery
            logging.error("%s does not exist", method_name)
            invocation.return_error_literal(Gio.dbus_error_quark(), Gio.DBusError.NOT_SUPPORTED, f"Invalid method {method_name}")

    def player_handle_get_property(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, value: Any):
        """
        Handles properties for org.mpris.MediaPlayer2.Player
        """
        assert interface_name == "org.mpris.MediaPlayer2.Player", f"Wrong interface name {interface_name} from {sender}"
        assert value in self.player_properties.keys(), f"{value} does not exist"

        return self.player_properties[value]

    def player_handle_set_property(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, key: str, value: Any) -> bool:
        """
        Handles properties for org.mpris.MediaPlayer2.Player
        """
        assert interface_name == "org.mpris.MediaPlayer2.Player", f"Wrong interface name {interface_name} from {sender}"
        assert key in self.player_properties.keys(), f"{key} does not exist"

        print(f"player_handle_set_property key: {key}, value: {value}", file=sys.stderr, flush=True)

        if key == "Rate":
            self.set_rate(value, connection, object_path)
        elif key == "LoopStatus":
            self.set_repeat(value, connection, object_path)
        elif key == "Shuffle":
            self.set_shuffle(value, connection, object_path)
        elif key == "Volume":
            self.set_volume(value, connection, object_path)
        else:
            return False

        # What is the correct thing to return here on success?  It appears that
        # we need to return something other than None or what would be evaluated
        # to False for this call back to be successful.
        return True

    def interface_handle_method_call(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, method_name: str, parameters: GLib.Variant, invocation: Gio.DBusMethodInvocation) -> None:
        """
        Handles method calls for org.mpris.MediaPlayer2
        """
        assert interface_name == "org.mpris.MediaPlayer2", f"Wrong interface name {interface_name} from {sender}"

        if method_name == "Raise":
            print("Raise", file=sys.stderr, flush=True)
        elif method_name == "Quit":
            print("Quit", file=sys.stderr, flush=True)
        else:
            logging.error("%s does not exist", method_name)
            invocation.return_error_literal(Gio.dbus_error_quark(), Gio.DBusError.NOT_SUPPORTED, f"Invalid method {method_name}")

    def interface_handle_get_property(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, value: Any):
        """
        Handles properties for org.mpris.MediaPlayer2
        """
        assert interface_name == "org.mpris.MediaPlayer2", f"Wrong interface name {interface_name} from {sender}"
        assert value in self.base_properties.keys(), f"{value} does not exist"

        return self.base_properties[value]

    def interface_handle_set_property(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, key: str, value: Any) -> bool:
        """
        Handles properties for org.mpris.MediaPlayer2
        """
        assert interface_name == "org.mpris.MediaPlayer2", f"Wrong interface name {interface_name} from {sender}"
        assert key in self.base_properties.keys(), f"{key} does not exist"

        return True


class InvalidMpris2(Mpris2):
    """
    MPRIS2 interfaces with invalid responses
    """

    APP_INTERFACE: Final = f"org.mpris.MediaPlayer2.appiumtest_invalid.instance{str(getpid())}"

    def __init__(self) -> None:
        super().__init__([], {}, {}, 0)

    def properties_handle_method_call(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, method_name: str, parameters: GLib.Variant, invocation: Gio.DBusMethodInvocation) -> None:
        """
        @override
        """
        invocation.return_error_literal(Gio.dbus_error_quark(), Gio.DBusError.NOT_SUPPORTED, "")

    def player_handle_method_call(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, method_name: str, parameters: GLib.Variant, invocation: Gio.DBusMethodInvocation) -> None:
        """
        @override
        """
        invocation.return_error_literal(Gio.dbus_error_quark(), Gio.DBusError.NOT_SUPPORTED, "")

    def player_handle_get_property(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, value: Any):
        """
        @override
        """
        return None

    def player_handle_set_property(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, key: str, value: Any) -> bool:
        """
        @override
        """
        return False

    def interface_handle_method_call(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, method_name: str, parameters: GLib.Variant, invocation: Gio.DBusMethodInvocation) -> None:
        """
        @override
        """
        invocation.return_error_literal(Gio.dbus_error_quark(), Gio.DBusError.NOT_SUPPORTED, "")

    def interface_handle_get_property(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, value: Any):
        """
        @override
        """
        return None

    def interface_handle_set_property(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, key: str, value: Any) -> bool:
        """
        @override
        """
        return False


player: Mpris2
loopThread: GLibMainLoopThread


def __on_terminate(signal, frame) -> None:
    player.quit()
    loopThread.quit()
    if loopThread.is_alive():
        loopThread.join(timeout=10)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="MPRIS2 media player for testing")
    parser.add_argument("json_path", nargs="?", help="Path to JSON file with player metadata")
    parser.add_argument("--invalid", action="store_true", help="Start an invalid player for error testing")
    parser.add_argument("--start-index", type=int, default=0, help="Starting track index")
    args = parser.parse_args()

    signal.signal(signal.SIGTERM, __on_terminate)

    loopThread = GLibMainLoopThread()
    loopThread.start()

    if args.invalid:
        player = InvalidMpris2()
        logging.info("Invalid player started")
    else:
        if not args.json_path:
            parser.error("json_path is required when not using --invalid")
        with open(args.json_path, "r", encoding="utf-8") as f:
            json_dict: dict[str, list | dict] = json.load(f)
        metadata: list[dict[str, GLib.Variant]] = read_player_metadata(json_dict)
        base_properties: dict[str, GLib.Variant] = read_base_properties(json_dict)
        current_index: int = args.start_index
        player_properties: dict[str, GLib.Variant] = read_player_properties(json_dict, metadata[current_index])
        player = Mpris2(metadata, base_properties, player_properties, current_index)
        logging.info("Player %s started", base_properties['Identity'].get_string())



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2025 Bohdan Onofriichuk <bogdan.onofriuchuk@gmail.com>
# SPDX-License-Identifier: MIT

"""udisks2 mock template

This creates the expected methods and properties of the main
UDisks2 object and the manager, and adds methods to manipulate a simple
mock drive.
"""

import dbus

from datetime import datetime

import dbusmock
from dbusmock import MOCK_IFACE

SYSTEM_BUS = True
IS_OBJECT_MANAGER = True

BUS_NAME = "org.freedesktop.UDisks2"
MAIN_OBJ = "/org/freedesktop/UDisks2"
MAIN_IFACE = 'org.freedesktop.UDisks2'

UDISKS2_MANAGER_IFACE = "org.freedesktop.UDisks2.Manager"
UDISKS2_MANAGER_OBJ = "/org/freedesktop/UDisks2/Manager"

UDISKS2_BLOCK_DEVICES = "/org/freedesktop/UDisks2/block_devices/"
UDISKS2_DRIVE_DEVICES = "/org/freedesktop/UDisks2/drives/"
UDISKS2_JOBS = "/org/freedesktop/UDisks2/jobs/"

UDISKS2_JOB_IFACE = "org.freedesktop.UDisks2.Job"
UDISKS2_DRIVE_IFACE = "org.freedesktop.UDisks2.Drive"
UDISKS2_BLOCK_IFACE = "org.freedesktop.UDisks2.Block"
UDISKS2_PARTITION_IFACE = "org.freedesktop.UDisks2.Partition"
UDISKS2_FILESYSTEM_IFACE = "org.freedesktop.UDisks2.Filesystem"

MOUNT_POINT = "/run/media/"

def load(mock, parameters):

    supported_filesystems = ["ext4"]

    manager_props = {
        "Version": dbus.String("2.10.2"),
        "SupportedFilesystems": dbus.Array(supported_filesystems, signature="s"),
        "SupportedEncryptionTypes": dbus.Array([], signature="s"),
        "DefaultEncryptionTypes": dbus.Array([], signature="s"),
    }

    manager_methods = [
        ("CanFormat", "s", "(bs)", "ret = self.can_format(args[0]), \"\""),
        ("CanResize", "s", "(bts)", "ret = self.can_resize(args[0]), \"\""),
        ("CanCheck", "s", "(bs)", "ret = self.can_check(args[0]), \"\""),
        ("CanRepair", "s", "(bs)", "ret = self.can_repair(args[0]), \"\""),
        ("GetBlockDevices", "a{sv}", "ao", "ret = self.get_block_paths()"),
        ("GetDrives", "a{sv}", "ao", "ret = self.get_drive_paths()"),
    ]

    mock.AddObject(UDISKS2_MANAGER_OBJ, UDISKS2_MANAGER_IFACE, manager_props, manager_methods)

    obj = dbusmock.get_object(UDISKS2_MANAGER_OBJ)
    obj.block_devices = []
    obj.drive_devices = []
    obj.dev_t = 0
    obj.job_counter = 0
    obj.supported_filesystems = supported_filesystems

@dbus.service.method(MOCK_IFACE,
                     in_signature="sttb", out_signature="")
def add_mock_device(self, device_name, device_number, device_size, is_damaged):
    """
    Convenience method to add a mock device
    """

    block_full_name = device_name + str(device_number)
    drive_full_name = "MOCK_FLASH_" + block_full_name

    drive_props = {
        "Vendor": dbus.String("MockDevice"),
        "Model": dbus.String("MockModel"),
        "Revision": dbus.String("MockRevision"),
        "Serial": dbus.String("MockSerial"),
        "WWN": dbus.String("MockWWN"),
        "Id": dbus.String("by-mock-" + str(device_number)),
        "Configuration": dbus.Array([], signature="(sa{sv})"),
        "Media": dbus.String(""),
        "MediaCompatibility": dbus.Array([], signature="s"),
        "MediaRemovable": dbus.Boolean(True),
        "MediaAvailable": dbus.Boolean(True),
        "MediaChangeDetected": dbus.Boolean(True),
        "Size": dbus.UInt64(device_size),
        "TimeDetected": dbus.UInt64(0),
        "TimeMediaDetected": dbus.UInt64(0),
        "Optical": dbus.Boolean(False),
        "OpticalBlank": dbus.Boolean(False),
        "OpticalNumTracks": dbus.UInt32(0),
        "OpticalNumAudioTracks": dbus.UInt32(0),
        "OpticalNumDataTracks": dbus.UInt32(0),
        "OpticalNumSessions": dbus.UInt32(0),
        "RotationRate": dbus.Int32(-1),
        "ConnectionBus": dbus.String("usb"),
        "Seat": dbus.String("seat0"),
        "Removable": dbus.Boolean(True),
        "Ejectable": dbus.Boolean(True),
        "SortKey": dbus.String("mock"),
        "CanPowerOff": dbus.Boolean(True),
        "SiblingId": dbus.String("mock")
    }

    drive_methods = [
        ("Eject", "a{sv}", "", "self.eject_drive(self.drive_name)"),
        ("SetConfiguration", "a{sv}a{sv}", "", "pass"),
        ("PowerOff", "a{sv}", "", "pass"),
    ]

    mng_obj = dbusmock.get_object(UDISKS2_MANAGER_OBJ)

    block_props = {
        "Device": dbus.ByteArray((MOUNT_POINT + block_full_name).encode("ascii")),
        "PreferredDevice": dbus.ByteArray((MOUNT_POINT + block_full_name).encode("ascii")),
        "Symlinks": dbus.Array([], signature="ay"),
        "DeviceNumber": dbus.UInt64(mng_obj.dev_t),
        "Id": dbus.String("by-mock-" + block_full_name),
        "Size": dbus.UInt64(device_size),
        "ReadOnly": dbus.Boolean(False),
        "Drive": dbus.ObjectPath(UDISKS2_DRIVE_DEVICES + drive_full_name),
        "MDRaid": dbus.ObjectPath("/"),
        "MDRaidMember": dbus.ObjectPath("/"),
        "IdUsage": dbus.String("filesystem"),
        "IdType": dbus.String("ext4"),
        "IdVersion": dbus.String("1.0"),
        "IdLabel": dbus.String(block_full_name),
        "IdUUID": dbus.String("86bf4da8-af29-406a-86f6-8eb9e11c99bd"),
        "Configuration": dbus.Array([], signature="(sa{sv})"),
        "CryptoBackingDevice": dbus.ObjectPath("/"),
        "HintPartitionable": dbus.Boolean(True),
        "HintSystem": dbus.Boolean(False),
        "HintIgnore": dbus.Boolean(False),
        "HintAuto": dbus.Boolean(False),
        "HintName": dbus.String(""),
        "HintIconName": dbus.String(""),
        "HintSymbolicIconName": dbus.String(""),
        "UserspaceMountOptions": dbus.Array([], signature="s"),
    }

    mng_obj.dev_t += 1

    block_methods = [
        ("AddConfigurationItem", "(sa{sv})a{sv}", "", "pass"),
        ("RemoveConfigurationItem", "(sa{sv})a{sv}", "", "pass"),
        ("UpdateConfigurationItem", "(sa{sv})(sa{sv})", "", "pass"),
        ("GetSecretInformation", "a{sv}", "a(sa{sv})", "pass"),
        ("Format", "sa{sv}", "", "pass"),
        ("OpenForBackup", "sa{sv}", "h", "pass"),
        ("OpenForRestore", "sa{sv}", "h", "pass"),
        ("OpenForBenchmark", "sa{sv}", "h", "pass"),
        ("OpenDevice", "sa{sv}", "h", "pass"),
        ("Rescan", "a{sv}", "", "pass"),
        ("RestoreEncryptedHeader", "sa{sv}", "", "pass"),
    ]

    filesystem_props = {
        "MountPoints": dbus.Array([], signature="ay"),
        "Size": dbus.UInt64(device_size),
    }

    filesystem_methods = [
        ("SetLabel", "sa{sv}", "", "pass"),
        ("SetUUID", "sa{sv}", "", "pass"),
        ("Mount", "a{sv}", "s", "ret = self.mount_filesystem(self.block_name)"),
        ("Unmount", "a{sv}", "", "self.unmount_filesystem(self.block_name)"),
        ("Resize", "ta{sv}", "", "pass"),
        ("Check", "a{sv}", "b", "ret = self.check_filesystem(self.block_name)"),
        ("Repair", "a{sv}", "b", "ret = self.repair_filesystem(self.block_name)"),
        ("TakeOwnership", "a{sv}", "", "pass"),
    ]

    block_device_path = UDISKS2_BLOCK_DEVICES + block_full_name
    drive_device_path = UDISKS2_DRIVE_DEVICES + drive_full_name

    self.save_block_path(block_device_path)
    self.save_drive_path(drive_device_path)

    self.AddObject(drive_device_path, UDISKS2_DRIVE_IFACE, drive_props, drive_methods)

    self.AddObject(block_device_path, UDISKS2_BLOCK_IFACE, block_props, block_methods)

    drv_obj = dbusmock.get_object(drive_device_path)
    drv_obj.block_name = block_full_name
    drv_obj.drive_name = drive_full_name

    blk_obj = dbusmock.get_object(block_device_path)
    blk_obj.block_name = block_full_name
    blk_obj.drive_name = drive_full_name
    blk_obj.is_damaged = is_damaged
    blk_obj.AddProperties(UDISKS2_FILESYSTEM_IFACE, filesystem_props)
    blk_obj.AddMethods(UDISKS2_FILESYSTEM_IFACE, filesystem_methods)

    main_obj = dbusmock.get_object(MAIN_OBJ)
    main_obj.object_manager_emit_added(drive_device_path)
    main_obj.object_manager_emit_added(block_device_path)

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="")
def remove_mock_device(self, device_name):
    """
    Convenience method to remove a mock device
    """

    block_path = UDISKS2_BLOCK_DEVICES + device_name

    blk_obj = dbusmock.get_object(block_path)

    drive_path = UDISKS2_DRIVE_DEVICES + blk_obj.drive_name

    self.object_manager_emit_removed(block_path)

    self.RemoveObject(block_path)

    self.remove_block_path(block_path)

    self.remove_drive_path(drive_path)

    self.object_manager_emit_removed(drive_path)

    self.RemoveObject(drive_path)

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="s")
def mount_filesystem(self, device_name):
    """
    Convenience method to mount a mock block device object
    simulates the org.freedesktop.UDisks2.Filesystem.Mount
    """

    path = UDISKS2_BLOCK_DEVICES + device_name

    job_path = self.create_job("filesystem-mount", path)

    self.finish_job(job_path)

    mount_point = MOUNT_POINT + device_name

    obj = dbusmock.get_object(path)

    mount_points = dbus.Array([], signature="ay")
    mount_points.append(dbus.ByteArray(mount_point.encode("ascii")))

    obj.Set(UDISKS2_FILESYSTEM_IFACE, "MountPoints", mount_points)

    userspace_mount_options = dbus.Array([], signature="s")
    userspace_mount_options.append(dbus.String("uhelper=udisks2"))

    obj.Set(UDISKS2_BLOCK_IFACE, "UserspaceMountOptions", userspace_mount_options)

    return path

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="")
def unmount_filesystem(self, device_name):
    """
    Convenience method to unmount a mock block device object
    simulate org.freedesktop.UDisks2.Filesystem.Unmount
    """

    path = UDISKS2_BLOCK_DEVICES + device_name

    job_path = self.create_job("filesystem-unmount", path)

    self.finish_job(job_path)

    obj = dbusmock.get_object(path)

    mount_points = dbus.Array([], signature="ay")

    obj.Set(UDISKS2_FILESYSTEM_IFACE, "MountPoints", mount_points)

    userspace_mount_options = dbus.Array([], signature="s")

    obj.Set(UDISKS2_BLOCK_IFACE, "UserspaceMountOptions", userspace_mount_options)

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="b")
def check_filesystem(self, device_name):
    """
    Convenience method to check a mock block device object
    simulates org.freedesktop.UDisks2.Filesystem.Check
    """

    path = UDISKS2_BLOCK_DEVICES + device_name

    job_path = self.create_job("filesystem-check", path)

    self.finish_job(job_path)

    obj = dbusmock.get_object(path)

    return not obj.is_damaged

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="b")
def repair_filesystem(self, device_name):
    """
    Convenience method to repair a mock block device object
    simulates org.freedesktop.UDisks2.Filesystem.Repair
    """

    path = UDISKS2_BLOCK_DEVICES + device_name

    job_path = self.create_job("filesystem-repair", path)

    obj = dbusmock.get_object(path)

    self.finish_job(job_path)

    obj.is_damaged = False

    return True

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="")
def eject_drive(self, drive_name):
    """
    Convenience method to eject a mock block device object which
    simulates org.freedesktop.UDisks2.Drive.Eject
    """

    drive_path = UDISKS2_DRIVE_DEVICES + drive_name

    drive_obj = dbusmock.get_object(drive_path)

    block_path = UDISKS2_BLOCK_DEVICES + drive_obj.block_name

    old_block_obj = dbusmock.get_object(block_path)

    # Workaround: python-dbusmock does not support removing of individual dbus ifaces.
    # So remove the block_devices object and then add a new object without the
    # 'org.freedesktop.UDisks2.Filesystem' iface.
    block_props = {
        "Device": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "Device"),
        "PreferredDevice": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "PreferredDevice"),
        "Symlinks": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "Symlinks"),
        "DeviceNumber": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "DeviceNumber"),
        "Id": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "Id"),
        "Size": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "Size"),
        "ReadOnly": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "ReadOnly"),
        "Drive": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "Drive"),
        "MDRaid": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "MDRaid"),
        "MDRaidMember": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "MDRaidMember"),
        "IdUsage": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "IdUsage"),
        "IdType": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "IdType"),
        "IdVersion": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "IdVersion"),
        "IdLabel": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "IdLabel"),
        "IdUUID": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "IdUUID"),
        "Configuration": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "Configuration"),
        "CryptoBackingDevice": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "CryptoBackingDevice"),
        "HintPartitionable": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "HintPartitionable"),
        "HintSystem": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "HintSystem"),
        "HintIgnore": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "HintIgnore"),
        "HintAuto": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "HintAuto"),
        "HintName": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "HintName"),
        "HintIconName": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "HintIconName"),
        "HintSymbolicIconName": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "HintSymbolicIconName"),
        "UserspaceMountOptions": old_block_obj.Get(UDISKS2_BLOCK_IFACE, "UserspaceMountOptions"),
    }

    block_methods = [
        ("AddConfigurationItem", "(sa{sv})a{sv}", "", "pass"),
        ("RemoveConfigurationItem", "(sa{sv})a{sv}", "", "pass"),
        ("UpdateConfigurationItem", "(sa{sv})(sa{sv})", "", "pass"),
        ("GetSecretInformation", "a{sv}", "a(sa{sv})", "pass"),
        ("Format", "sa{sv}", "", "pass"),
        ("OpenForBackup", "sa{sv}", "h", "pass"),
        ("OpenForRestore", "sa{sv}", "h", "pass"),
        ("OpenForBenchmark", "sa{sv}", "h", "pass"),
        ("OpenDevice", "sa{sv}", "h", "pass"),
        ("Rescan", "a{sv}", "", "pass"),
        ("RestoreEncryptedHeader", "sa{sv}", "", "pass"),
    ]

    changed_block_props = {
        "HintAuto": dbus.Boolean(0),
        "IdUUID": dbus.String(""),
        "IdVersion": dbus.String(""),
        "IdType": dbus.String(""),
        "IdUsage": dbus.String(""),
        "Id": dbus.String(""),
        "Size": dbus.UInt64(0),
    }

    changed_drive_props = {
        "TimeMediaDetected": dbus.UInt64(0),
        "MediaAvailable": dbus.Boolean(False),
        "Size": dbus.UInt64(0),
    }

    job_path = self.create_job("drive-eject", drive_path)

    self.finish_job(job_path)

    self.RemoveObject(block_path)

    self.AddObject(block_path, UDISKS2_BLOCK_IFACE, block_props, block_methods)

    main_obj = dbusmock.get_object(MAIN_OBJ)

    drive_obj.UpdateProperties(UDISKS2_DRIVE_IFACE, changed_drive_props)

    new_block_obj = dbusmock.get_object(block_path)
    new_block_obj.block_name = old_block_obj.block_name
    new_block_obj.drive_name = old_block_obj.drive_name
    new_block_obj.is_damaged = old_block_obj.is_damaged
    new_block_obj.UpdateProperties(UDISKS2_BLOCK_IFACE, changed_block_props)

    main_obj.EmitSignal(
        dbusmock.OBJECT_MANAGER_IFACE, "InterfacesRemoved", "oas", [dbus.ObjectPath(block_path), [dbus.String(UDISKS2_FILESYSTEM_IFACE)]]
    )

@dbus.service.method(MOCK_IFACE,
                     in_signature="ss", out_signature="s")
def create_job(self, operation, device_path) -> str:
    """
    Convenience method to create a job
    simulates org.freedesktop.UDisks2.Job
    """

    mng_obj = dbusmock.get_object(UDISKS2_MANAGER_OBJ)

    job_path = UDISKS2_JOBS + str(mng_obj.job_counter)
    mng_obj.job_counter += 1

    dt = datetime.now()

    props = {
        "Operation": dbus.String(operation),
        "Progress": dbus.Double(0),
        "ProgressValid": dbus.Boolean(False),
        "Bytes": dbus.UInt64(0),
        "Rate": dbus.UInt64(0),
        "StartTime": dbus.UInt64(dt.timestamp()),
        "ExpectedEndTime": dbus.UInt64(0),
        "Objects": dbus.Array([device_path], signature="o"),
        "StartedByUID": dbus.UInt32(0),
        "Cancelable": dbus.Boolean(False),
    }

    methods = [
        ("Cancel", "a{sv}", "", "pass"),
    ]

    self.AddObject(job_path, UDISKS2_JOB_IFACE, props, methods)

    self.object_manager_emit_added(job_path)

    return job_path

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="")
def finish_job(self, job_path):
    """
    Convenience method to remove a job
    simulates org.freedesktop.UDisks2.Job
    """

    obj = dbusmock.get_object(job_path)

    obj.EmitSignal(UDISKS2_JOB_IFACE, "Completed", "bs", [True,""])

    self.object_manager_emit_removed(job_path)

    self.RemoveObject(job_path)

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="b")
def can_format(_self, filesystem):
    #TODO: implement
    return False, 0

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="(bt)")
def can_resize(_self, filesystem):
    #TODO: implement
    return False, 0

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="b")
def can_check(_self, filesystem):
    """
    Convenience method to check if filesystem can be checked
    simulates org.freedesktop.UDisks2.Manager.CanCheck
    """

    obj = dbusmock.get_object(UDISKS2_MANAGER_OBJ)
    return any(elem in filesystem for elem in obj.supported_filesystems)

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="b")
def can_repair(_self, filesystem):
    """
    Convenience method to check if filesystem can be repaired
    simulates org.freedesktop.UDisks2.Manager.CanRepair
    """

    obj = dbusmock.get_object(UDISKS2_MANAGER_OBJ)
    return any(elem in filesystem for elem in obj.supported_filesystems)

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="")
def save_block_path(_self, path):
    """
    Convenience method to maintain a list of block devices which is used by
    org.freedesktop.UDisks2.Manager.GetBlockDevices
    """

    obj = dbusmock.get_object(UDISKS2_MANAGER_OBJ)
    obj.block_devices.append(str(path))

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="")
def remove_block_path(_self, path):
    """
    Convenience method to maintain a list of block devices which is used by
    org.freedesktop.UDisks2.Manager.GetBlockDevices
    """

    obj = dbusmock.get_object(UDISKS2_MANAGER_OBJ)
    obj.block_devices.remove(str(path))

@dbus.service.method(MOCK_IFACE,
                     in_signature="", out_signature="ao")
def get_block_paths(_self) -> list[str]:
    """
    Convenience method to maintain a list of block devices which is used by
    org.freedesktop.UDisks2.Manager.GetBlockDevices
    """

    obj = dbusmock.get_object(UDISKS2_MANAGER_OBJ)
    return obj.block_devices

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="")
def save_drive_path(_self, path):
    """
    Convenience method to maintain a list of drive devices which is used by
    org.freedesktop.UDisks2.Manager.GetDrives
    """

    obj = dbusmock.get_object(UDISKS2_MANAGER_OBJ)
    obj.drive_devices.append(str(path))

@dbus.service.method(MOCK_IFACE,
                     in_signature="s", out_signature="")
def remove_drive_path(_self, path):
    """
    Convenience method to maintain a list of drive devices which is used by
    org.freedesktop.UDisks2.Manager.GetDrives
    """

    obj = dbusmock.get_object(UDISKS2_MANAGER_OBJ)
    obj.drive_devices.remove(str(path))

@dbus.service.method(MOCK_IFACE,
                     in_signature="", out_signature="ao")
def get_drive_paths(_self) -> list[str]:
    """
    Convenience method to maintain a list of drive devices which is used by
    org.freedesktop.UDisks2.Manager.GetDrives
    """

    obj = dbusmock.get_object(UDISKS2_MANAGER_OBJ)
    return obj.drive_devices



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Fushan Wen <qydwhotmail@gmail.com>
# SPDX-License-Identifier: MIT

import os
import subprocess
import sys
import threading
from typing import Final

from gi.repository import Gio, GLib

CMAKE_BINARY_DIR: Final = os.getenv("CMAKE_BINARY_DIR")
QMLTEST_PATH: Final = os.getenv("QMLTEST_PATH")


class GLibMainLoopThread(threading.Thread):

    def __init__(self) -> None:
        # Set up D-Bus loop
        self.loop = GLib.MainLoop()
        self.failSafeTimer = threading.Timer(600, self.loop.quit)

        # Create the thread
        super().__init__()

    def run(self) -> None:
        self.failSafeTimer.start()
        self.loop.run()

    def quit(self) -> None:
        self.failSafeTimer.cancel()
        self.loop.quit()


class TestInterface:

    BUS_NAME: Final = "org.kde.mpristest"
    OBJECT_PATH: Final = "/mpristest"

    connection: Gio.DBusConnection
    pid: int

    def __init__(self) -> None:
        self.reg_id: int = 0
        self.registered_event = threading.Event()
        self.owner_id: int = Gio.bus_own_name(Gio.BusType.SESSION, self.BUS_NAME, Gio.BusNameOwnerFlags.NONE, self.on_bus_acquired, None, None)
        assert self.owner_id > 0

    def quit(self) -> None:
        self.connection.unregister_object(self.reg_id)
        Gio.bus_unown_name(self.owner_id)
        self.connection.flush_sync(None)

    def on_bus_acquired(self, connection: Gio.DBusConnection, name: str, *args) -> None:
        """
        Interface is ready, now register objects.
        """
        self.connection = connection

        introspection_data = Gio.DBusNodeInfo.new_for_xml("""
<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
    "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
<node>
  <interface name="org.kde.mpristest">
    <method name="GetPid">
      <arg name="outArg" type="i" direction="out"/>
    </method>
  </interface>
</node>
""")
        self.reg_id = connection.register_object(self.OBJECT_PATH, introspection_data.interfaces[0], self.handle_method_call, None, None)
        assert self.reg_id > 0

        self.registered_event.set()

    def handle_method_call(self, connection: Gio.DBusConnection, sender: str, object_path: str, interface_name: str, method_name: str, parameters: GLib.Variant, invocation: Gio.DBusMethodInvocation) -> None:
        if method_name == "GetPid":
            invocation.return_value(GLib.Variant("(i)", [self.pid]))
        else:
            invocation.return_error_literal(Gio.dbus_error_quark(), Gio.DBusError.UNKNOWN_METHOD, f"Unknown method {method_name}")


if __name__ == '__main__':
    assert QMLTEST_PATH is not None
    assert os.path.exists(QMLTEST_PATH)
    result: int = 1

    loop = GLibMainLoopThread()
    loop.start()
    dbus_interface = TestInterface()

    player_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, os.pardir, "appiumtests", "applets", "mediacontrollertest")
    assert os.path.exists(player_dir)
    with subprocess.Popen(["python3", os.path.join(player_dir, "mediaplayer.py"), os.path.join(player_dir, "player_b.json")], stdout=sys.stdout, stderr=sys.stderr) as player:
        TestInterface.pid = player.pid
        args = [
            QMLTEST_PATH,
            "-input",
            os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", sys.argv[1]),
        ]
        if CMAKE_BINARY_DIR is not None:
            args += ["-import", CMAKE_BINARY_DIR]

        print("Starting test")
        test_runner = subprocess.Popen(args, stdout=sys.stdout, stderr=sys.stderr)
        try:
            result = test_runner.wait(timeout=60)
        except subprocess.TimeoutExpired as e:
            print("Timeout", e)
        finally:
            test_runner.terminate()
            player.terminate()
            test_runner.wait(10)

    dbus_interface.quit()
    loop.quit()

    sys.exit(result)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2023 Fushan Wen <qydwhotmail@gmail.com>
# SPDX-License-Identifier: MIT

import os
import sqlite3
import subprocess
import sys
import tempfile
import time
import unittest
from typing import Final

import gi

gi.require_version('GdkPixbuf', '2.0')
from gi.repository import GdkPixbuf

CMAKE_RUNTIME_OUTPUT_DIRECTORY: Final = os.getenv("CMAKE_RUNTIME_OUTPUT_DIRECTORY", os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, os.pardir, "build", "bin"))


class V3MigrationTest(unittest.TestCase):
    """
    history2.lst -> history3.sqlite
    """

    v2_to_v3_bin: str
    history2_text_file: str
    history2_image_file: str
    time_now: int = int(time.time())

    @classmethod
    def setUpClass(cls) -> None:
        cls.v2_to_v3_bin = os.path.join(CMAKE_RUNTIME_OUTPUT_DIRECTORY, "plasma6.3-update-clipboard-database-2-to-3")
        cls.history2_text_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "onetextentry.lst")
        cls.history2_image_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "bug465225.lst")
        assert os.path.exists(cls.v2_to_v3_bin)
        assert os.path.exists(cls.history2_text_file)
        assert os.path.exists(cls.history2_image_file)

    @classmethod
    def tearDownClass(cls) -> None:
        pass

    def test_convert_v2_to_v3_text(self) -> None:

        with tempfile.TemporaryDirectory() as temp_dir:
            target_db = os.path.join(temp_dir, "history3.sqlite")
            subprocess.check_call([self.v2_to_v3_bin, "--input", self.history2_text_file, "--output", target_db], stdout=sys.stdout, stderr=sys.stderr)
            self.assertTrue(os.path.exists(os.path.join(temp_dir, "data")))
            self.assertTrue(os.path.exists(target_db))
            for uuid, result in (("230aa750d982a8e1a7e8f0b7ccc4e4b1b87bf593", "Fushan Wen"), ("e2ab8561c5a8f9967e62486c44211c63bcf7d002", "clipboard")):
                self.assertTrue(os.path.exists(os.path.join(temp_dir, "data", uuid)))
                self.assertTrue(os.path.exists(os.path.join(temp_dir, "data", uuid, uuid)))
                with open(os.path.join(temp_dir, "data", uuid, uuid), encoding="utf-8") as fh:
                    self.assertEqual(fh.readline(), result)

            con = sqlite3.connect(target_db)
            cur = con.cursor()
            res_cur = cur.execute("SELECT uuid,added_time,mimetypes,text FROM main ORDER BY added_time DESC")

            res = res_cur.fetchone()
            self.assertEqual(res[0], "230aa750d982a8e1a7e8f0b7ccc4e4b1b87bf593")
            self.assertGreaterEqual(res[1], self.time_now)
            self.assertEqual(res[2], "text/plain,text/plain;charset=utf-8")
            self.assertEqual(res[3], "Fushan Wen")

            # main
            res = res_cur.fetchone()
            self.assertEqual(res[0], "e2ab8561c5a8f9967e62486c44211c63bcf7d002")
            self.assertGreaterEqual(res[1], self.time_now - 1)
            self.assertEqual(res[2], "text/plain,text/plain;charset=utf-8")
            self.assertEqual(res[3], "clipboard")

            # aux
            self.assertEqual(1, len(cur.execute("SELECT uuid,mimetype,data_uuid FROM aux WHERE uuid='230aa750d982a8e1a7e8f0b7ccc4e4b1b87bf593' AND mimetype='text/plain;charset=utf-8' AND data_uuid='230aa750d982a8e1a7e8f0b7ccc4e4b1b87bf593'").fetchall()))
            self.assertEqual(1, len(cur.execute("SELECT uuid,mimetype,data_uuid FROM aux WHERE uuid='e2ab8561c5a8f9967e62486c44211c63bcf7d002' AND mimetype='text/plain;charset=utf-8' AND data_uuid='e2ab8561c5a8f9967e62486c44211c63bcf7d002'").fetchall()))

    def test_convert_v2_to_v3_image(self) -> None:
        with tempfile.TemporaryDirectory() as temp_dir:
            target_db = os.path.join(temp_dir, "history3.sqlite")
            subprocess.check_call([self.v2_to_v3_bin, "--input", self.history2_image_file, "--output", target_db], stdout=sys.stdout, stderr=sys.stderr)
            self.assertTrue(os.path.exists(target_db))
            self.assertTrue(os.path.exists(os.path.join(temp_dir, "data")))
            uuid = "8f9353dabfdcf9aca5a901cd2c4ae6717cac5adc"
            self.assertTrue(os.path.exists(os.path.join(temp_dir, "data", uuid)))
            self.assertTrue(os.path.exists(os.path.join(temp_dir, "data", uuid, uuid)))
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(os.path.join(temp_dir, "data", uuid, uuid))
            self.assertIsNotNone(pixbuf)
            self.assertEqual(pixbuf.get_width(), 1610)
            self.assertEqual(pixbuf.get_height(), 1329)

            con = sqlite3.connect(target_db)
            cur = con.cursor()
            res = cur.execute("SELECT uuid,added_time,mimetypes,text FROM main ORDER BY last_used_time DESC, added_time DESC").fetchall()
            self.assertEqual(len(res), 1)

            # main
            self.assertEqual(res[0][0], "8f9353dabfdcf9aca5a901cd2c4ae6717cac5adc")
            self.assertGreaterEqual(res[0][1], self.time_now)
            self.assertEqual(res[0][2], "image/png")
            self.assertIsNone(res[0][3])

            # aux
            self.assertEqual(1, len(cur.execute("SELECT uuid,mimetype,data_uuid FROM aux WHERE uuid='8f9353dabfdcf9aca5a901cd2c4ae6717cac5adc' AND mimetype='image/png' AND data_uuid='8f9353dabfdcf9aca5a901cd2c4ae6717cac5adc'").fetchall()))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2023 Fushan Wen <qydwhotmail@gmail.com>
# SPDX-License-Identifier: MIT

import os
import subprocess
import sys
import time
import unittest
from typing import Final

from appium import webdriver
from appium.options.common.base import AppiumOptions
from appium.webdriver.common.appiumby import AppiumBy
from gi.repository import Gio, GLib
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

WIDGET_ID: Final = "org.kde.plasma.appmenu"
KDE_VERSION: Final = 6
CMAKE_RUNTIME_OUTPUT_DIRECTORY: Final = os.getenv("CMAKE_RUNTIME_OUTPUT_DIRECTORY", os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, os.pardir, "build", "bin"))


def name_has_owner(session_bus: Gio.DBusConnection, name: str) -> bool:
    """
    Whether the given name is available on session bus
    """
    message: Gio.DBusMessage = Gio.DBusMessage.new_method_call("org.freedesktop.DBus", "/", "org.freedesktop.DBus", "NameHasOwner")
    message.set_body(GLib.Variant("(s)", [name]))
    reply, _ = session_bus.send_message_with_reply_sync(message, Gio.DBusSendMessageFlags.NONE, 1000)
    return reply and reply.get_signature() == 'b' and reply.get_body().get_child_value(0).get_boolean()


class AppMenuTest(unittest.TestCase):
    """
    Tests for the appmenu widget
    """

    driver: webdriver.Remote
    kded: subprocess.Popen | None = None

    @classmethod
    def setUpClass(cls) -> None:
        """
        Opens the widget and initialize the webdriver
        """
        session_bus: Gio.DBusConnection = Gio.bus_get_sync(Gio.BusType.SESSION)
        if not name_has_owner(session_bus, f"org.kde.kded{KDE_VERSION}"):
            cls.kded = subprocess.Popen([f"kded{KDE_VERSION}"])
            kded_started: bool = False
            for _ in range(10):
                if name_has_owner(session_bus, f"org.kde.kded{KDE_VERSION}"):
                    kded_started = True
                    break
                print(f"waiting for kded{KDE_VERSION} to appear on the dbus session")
                time.sleep(1)
            assert kded_started, "kded is not started"
            kded_reply: GLib.Variant = session_bus.call_sync(f"org.kde.kded{KDE_VERSION}", "/kded", f"org.kde.kded{KDE_VERSION}", "loadModule", GLib.Variant("(s)", ["appmenu"]), GLib.VariantType("(b)"), Gio.DBusSendMessageFlags.NONE, 1000)
            assert kded_reply.get_child_value(0).get_boolean(), "appmenu module is not loaded"

        options = AppiumOptions()
        options.set_capability("app", f"plasmawindowed -p org.kde.plasma.nano {WIDGET_ID}")
        options.set_capability("environ", {
            "LC_ALL": "en_US.UTF-8",
            "QT_LOGGING_RULES": "qt.accessibility.atspi.warning=false;qt.qml.propertyCache.append.warning=false",
        })
        options.set_capability("timeouts", {'implicit': 10000})
        cls.driver = webdriver.Remote(command_executor='http://127.0.0.1:4723', options=options)

    def tearDown(self) -> None:
        """
        Take screenshot when the current test fails
        """
        if not self._outcome.result.wasSuccessful():
            if os.getenv("TEST_WITH_KWIN_WAYLAND", "1") == "0":
                subprocess.check_call(["import", "-window", "root", f"failed_test_shot_{WIDGET_ID}_#{self.id()}.png"])
            else:
                self.driver.get_screenshot_as_file(f"failed_test_shot_{WIDGET_ID}_#{self.id()}.png")

    @classmethod
    def tearDownClass(cls) -> None:
        """
        Make sure to terminate the driver again, lest it dangles.
        """
        if cls.kded is not None:
            subprocess.check_call([f"kquitapp{KDE_VERSION}", f"kded{KDE_VERSION}"])
            cls.kded.wait(5)
        cls.driver.quit()

    def test_0_open(self) -> None:
        """
        Tests the widget can be opened
        """
        self.driver.find_element(AppiumBy.NAME, "Global Menu")

    def test_1_gtk_application_menu(self) -> None:
        """
        Activate a window with two hidden submenus and match strings in the widget
        """
        app = subprocess.Popen(["python3", os.path.join(os.path.dirname(os.path.abspath(__file__)), "appmenutest_window.py")], stdout=sys.stderr, stderr=sys.stderr)
        self.addCleanup(app.kill)

        menu1 = self.driver.find_element(AppiumBy.NAME, "foo")
        menu2 = self.driver.find_element(AppiumBy.NAME, "bar")

        app.terminate()
        app.wait()
        wait = WebDriverWait(self.driver, 5)
        wait.until(EC.invisibility_of_element(menu1))
        wait.until(EC.invisibility_of_element(menu2))


if __name__ == '__main__':
    try:
        from PySide6.QtCore import QObject
    except ModuleNotFoundError:
        sys.exit(0)
    unittest.main()
